/*
	Motorola Milestone overclock module
	version 1.2t - 2010-10-13
	by Tiago Sousa <mirage@kaotik.org>, modified by nadlabak
	License: GNU GPLv2
	<http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>

	http://code.google.com/p/milestone-overclock/

	Changelog:

        version 1.2t - 2010-10-13
        - adapted to the 2.6.32 Milestone test kernel (nadlabak)
	version 1.2 - 2010-05-25
	- change the values of freq_table by writing "index frequency" to
	  /proc/overclock/freq_table such as:
	  echo "1 400000" > /proc/overclock/freq_table
	- change the values of mpu_opps by writing "index rate vsel" to
	  /proc/overclock/mpu_opps such as:
	  echo "3 400000000 50" > /proc/overclock/mpu_opps
	- autodetect freq_table and mpu_opps given
	  omap2_clk_init_cpufreq_table's address at load time or:
	  echo 0xc004e498 > /proc/overclock/omap2_clk_init_cpufreq_table_addr
	  you can find the address with:
	  grep omap2_clk_init_cpufreq_table /proc/kallsyms
	version 1.1 - 2010-05-20
	- mpu_opps now configurable in load time
	- improved mpu_opps address lookup
	- corrected vsel units
	- applied mattih's patch to set policy->user_policy.max
	version 1.0 - 2010-05-14
	- initial release

	Description:

	The MPU (Microprocessor Unit) clock has 5 discrete pairs of possible
	rate frequencies and respective voltages, of which only 4 are passed
	down to cpufreq as you can see with a tool such as SetCPU.  The
	default frequencies are 125, 250, 500 and 550 MHz (and a hidden
	600).  By using this module, you are changing the highest pair in
	the tables of both cpufreq and MPU frequencies, so it becomes 125,
	250, 500 and, say, 800.  It's quite stable up to 1200; beyond
	that it quickly becomes unusable, specially over 1300, with lockups
	or spontaneous reboots.

	This version was prepared for Motorola Milestone's official Android
	2.1 for Central Europe, build number SHOLS_U2_02.31.0.  Fortunately
	Motorola appears to have used the same kernel in most of the 2.1
	firmwares.  Confirmed working are Central Europe, Telus and
	Brazilian firmwares with default addresses.  Milestones UK/FR with
	Android 2.0.1 can use freq_table_addr=0xc0509b24 and
	mpu_opps_addr=0xc050a848, Russian and Thai users currently need to
	change the kernel or use autodetect, see KernelModule wiki page at
	the project site for more information.  To port the module to
	different kernels or even other phones, see the Disassembly wiki
	page.  Testers welcome!

	Usage:

	insmod overclock.ko
	echo 62 > /proc/overclock/max_vsel
	echo 800000 > /proc/overclock/max_rate
	
	You should set max_vsel before max_rate if the new rate is going to
	be higher than the current one, because higher frequencies often
	require more voltage than supplied by default.  Likewise, lower the
	max_rate first before max_vsel if you want to reduce both frequency
	and voltage:
	echo 550000 > /proc/overclock/max_rate
	echo 56 > /proc/overclock/max_vsel

	To set a specified frequency and voltage at load time:
	insmod overclock.ko max_rate=800000 max_vsel=62

	Be careful when you try to load this in mot_boot_mode, it has been
	reported that the vsel is reset to default while rate stays
	overclocked.  Try to load it later in the boot process, for example
	with the Autostart app.

	Remember that you are merely changing the maximum possible value
	that cpufreq can choose to use.  The current speed may well be lower
	than the one specified if the phone is idle.  I recommend the use of
	the SetCPU app to effectively change the current frequency through
	its policies (use Autodetect Speeds instead of Droid/Milestone
	profile in the device settings).

	You may also supply the addresses of the two needed in-memory kernel
	structures which may change with the firmware build and will
	probably have to be supplied either at load time with the
	freq_table_addr and mpu_opps_addr parameters or with their
	respective /proc entries.  You can ask the module to try to
	autodetect the values by specifying the parameter
	omap2_clk_init_cpufreq_table_addr=0xaddress, where address can be
	retrieved from /proc/kallsyms.  Otherwise, finding out the values
	will require live disassembly of kernel code.  See the project page
	for more information.
*/

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>

#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#include <linux/kallsyms.h>

#include <linux/notifier.h>
#include <linux/cpufreq.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/cpu.h>
#include <linux/completion.h>
#include <linux/mutex.h>

#include <linux/err.h>
#include <linux/clk.h>
#include <linux/io.h>

#include <mach/hardware.h>
#include <asm/system.h>     
#include <plat/omap-pm.h>

#define DRIVER_AUTHOR "Tiago Sousa <mirage@kaotik.org>"
#define DRIVER_DESCRIPTION "Motorola Milestone CPU overclocking"
#define DRIVER_VERSION "1.2t (nadlabak's patch)"

#ifdef DROIDX
#define DRIVER_DEFAULT_RATE 1000000
#define DRIVER_DEFAULT_VSEL 56
#define DRIVER_DEFAULT_FREQ_TABLE_ADDR 0xc05a6368
#define DRIVER_DEFAULT_MPU_OPPS_ADDR 0xc0048a74
#define DRIVER_DEFAULT_CPUFREQ_ADDR 0xc00489e0
#elif DROID2
#define DRIVER_DEFAULT_RATE 1000000
#define DRIVER_DEFAULT_VSEL 56
#define DRIVER_DEFAULT_FREQ_TABLE_ADDR 0xc05a6368
#define DRIVER_DEFAULT_MPU_OPPS_ADDR 0xc0048a50
#define DRIVER_DEFAULT_CPUFREQ_ADDR 0xc00489bc
#else
#define DRIVER_DEFAULT_RATE 550000
#define DRIVER_DEFAULT_VSEL 56
#define DRIVER_DEFAULT_FREQ_TABLE_ADDR 0xc051794c
#define DRIVER_DEFAULT_MPU_OPPS_ADDR 0xc0518334
#define DRIVER_DEFAULT_CPUFREQ_ADDR 0
#endif

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("GPL");

static uint max_rate = DRIVER_DEFAULT_RATE;
static uint max_vsel = DRIVER_DEFAULT_VSEL;
static uint freq_table_addr = DRIVER_DEFAULT_FREQ_TABLE_ADDR;
static uint mpu_opps_addr = DRIVER_DEFAULT_MPU_OPPS_ADDR;
static uint omap2_clk_init_cpufreq_table_addr = DRIVER_DEFAULT_CPUFREQ_ADDR;

module_param(max_rate, uint, 0444);
module_param(max_vsel, uint, 0444);
module_param(freq_table_addr, uint, 0444);
module_param(mpu_opps_addr, uint, 0444);
#ifdef DROIDX
MODULE_PARM_DESC(max_rate, "The maximum MPU clock rate frequency in KHz (default 1000000)");
MODULE_PARM_DESC(max_vsel, "The maximum MPU voltage selection (default 56)");
MODULE_PARM_DESC(freq_table_addr, "The freq_table address (default 0xc00489e0)");
MODULE_PARM_DESC(mpu_opps_addr, "The mpu_opps address (default 0xc0048a74)");
#elif DROID2
MODULE_PARM_DESC(max_rate, "The maximum MPU clock rate frequency in KHz (default 1000000)");
MODULE_PARM_DESC(max_vsel, "The maximum MPU voltage selection (default 56)");
MODULE_PARM_DESC(freq_table_addr, "The freq_table address (default 0xc00489bc)");
MODULE_PARM_DESC(mpu_opps_addr, "The mpu_opps address (default 0xc0048a50)");
#else
MODULE_PARM_DESC(max_rate, "The maximum MPU clock rate frequency in KHz (default 550000)");
MODULE_PARM_DESC(max_vsel, "The maximum MPU voltage selection (default 56)");
MODULE_PARM_DESC(freq_table_addr, "The freq_table address (default 0xc051794c)");
MODULE_PARM_DESC(mpu_opps_addr, "The mpu_opps address (default 0xc0518334)");
#endif
module_param(omap2_clk_init_cpufreq_table_addr, uint, 0444);
MODULE_PARM_DESC(omap2_clk_init_cpufreq_table_addr, "The omap2_clk_init_cpufreq_table function address to autodetect the other addresses (default 0, doesn't autodetect)");

/* nadlabak: The structure must correspond with the structure used by the particular kernel.
 At this time I have no idea about the purpose of the data added by Motorola in the recent kernels.
 When released, the kernel sources will tell us more. */
struct omap_opp2 {
	unsigned long rate;
	u16 opp_id;
	u16 vseld;
        u16 vsel;
	u8 x0;
	u8 x1;
	u8 x2;
	u8 x3;
	u8 x4;
	u8 x5;
	u8 x6;
	u8 x7;
	u8 x8;
	u8 x9;
};

static struct cpufreq_frequency_table *freq_table;
static struct omap_opp2 *my_mpu_opps;
static struct cpufreq_policy *policy;
#define MPU_CLK         "arm_fck"
static struct clk *mpu_clk;

#define BUF_SIZE PAGE_SIZE
static char *buf;

static void set_max_speed(void)
{
	printk(KERN_INFO "overclock: setting max_rate %u and max_vsel %u\n", max_rate, max_vsel);
	freq_table[0].frequency = policy->max = policy->cpuinfo.max_freq =
		policy->user_policy.max = max_rate;
	my_mpu_opps[MAX_VDD1_OPP].vsel = max_vsel;
	my_mpu_opps[MAX_VDD1_OPP].rate = max_rate*1000;
}

static void omap2_find_addrs(void)
{
	unsigned char *func = (void *)omap2_clk_init_cpufreq_table_addr;
	uint *addr;
	int first = 1, i;
	for(i = 0; i < 100; i+=4) {
		if((func[i+3] == 0xe5 || func[i+3] == 0x15) /* ldr/ldrne */
			&& func[i+2] == 0x9f) { /* [pc, */
			addr = (void *)((uint)func)+i+8+func[i];
			if(first--) {
				mpu_opps_addr = *addr;
				printk (KERN_INFO "overclock: found mpu_opps_addr at 0x%x\n", mpu_opps_addr);
			} else {
				freq_table_addr = *addr;
				printk (KERN_INFO "overclock: found freq_table_addr at 0x%x\n", freq_table_addr);
				break;
			}
		}
	}
}

static int proc_info_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "cpumin=%u cpumax=%u min=%u max=%u usermin=%u usermax=%u\nclk_get_rate=%lu\n",
				policy->cpuinfo.min_freq, policy->cpuinfo.max_freq, policy->min, policy->max, policy->user_policy.min, policy->user_policy.max, clk_get_rate(mpu_clk) / 1000);

	return ret;
}

static int proc_max_rate_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "%u\n", max_rate);

	return ret;
}

static int proc_max_rate_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	ulong newrate;
	int result;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if((result = strict_strtoul(buf, 0, &newrate)))
		return result;
	if(max_rate != newrate) {
		max_rate = newrate;
		set_max_speed();
	}

	return len;
}

static int proc_max_vsel_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "%u\n", max_vsel);

	return ret;
}

static int proc_max_vsel_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	ulong newvsel;
	int result;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if((result = strict_strtoul(buf, 0, &newvsel)))
		return result;
	if(max_vsel != newvsel) {
		max_vsel = newvsel;
		set_max_speed();
	}

	return len;
}

static int proc_freq_table_addr_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "0x%x\n", (uint)freq_table);

	return ret;
}

static int proc_freq_table_addr_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	ulong newaddr;
	int result;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if((result = strict_strtoul(buf, 0, &newaddr)))
		return result;
	freq_table_addr = newaddr;
	freq_table = (void *)freq_table_addr;

	return len;
}

static int proc_mpu_opps_addr_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "0x%x\n", (uint)mpu_opps_addr);

	return ret;
}

static int proc_mpu_opps_addr_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	ulong newaddr;
	int result;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if((result = strict_strtoul(buf, 0, &newaddr)))
		return result;
	mpu_opps_addr = newaddr;
	my_mpu_opps = *(struct omap_opp2 **)mpu_opps_addr;

	return len;
}

static int proc_omap2_clk_init_cpufreq_table_addr_read(char *buffer,
		char **buffer_location, off_t offset, int count, int *eof,
		void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "0x%x\n",
			(uint)omap2_clk_init_cpufreq_table_addr);

	return ret;
}

static int proc_omap2_clk_init_cpufreq_table_addr_write(struct file *filp,
		const char __user *buffer, unsigned long len, void *data)
{
	ulong newaddr;
	int result;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if((result = strict_strtoul(buf, 0, &newaddr)))
		return result;
	omap2_clk_init_cpufreq_table_addr = newaddr;
	if(omap2_clk_init_cpufreq_table_addr)
		omap2_find_addrs();
	freq_table = (void *)freq_table_addr;
	my_mpu_opps = *(struct omap_opp2 **)mpu_opps_addr;

	return len;
}

static int proc_freq_table_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int i, ret = 0;
	
	if (offset > 0)
		ret = 0;
	else
		for(i = 0; freq_table[i].frequency != CPUFREQ_TABLE_END; i++) {
			if(ret >= count)
				break;
			ret += scnprintf(buffer+ret, count-ret, "freq_table[%d] index=%u frequency=%u\n", i, freq_table[i].index, freq_table[i].frequency);
		}

	return ret;
}

static int proc_freq_table_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	uint index, frequency;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if(sscanf(buf, "%d %d", &index, &frequency) == 2) {
		freq_table[index].frequency = frequency;
	} else
		printk(KERN_INFO "overclock: insufficient parameters for freq_table\n");

	return len;
}                        
                        
static int proc_mpu_opps_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int i, ret = 0;
	
	if (offset > 0)
		ret = 0;
	else
		for(i = MAX_VDD1_OPP; my_mpu_opps[i].rate; i--) {
			if(ret >= count)
				break;
// copy paste laziness:
			ret += scnprintf(buffer+ret, count-ret, "mpu_opps[%d] rate=%lu opp_id=%u vseld=%u vsel=%u %u %u %u %u %u %u %u %u %u %u\n", i, my_mpu_opps[i].rate, my_mpu_opps[i].opp_id, my_mpu_opps[i].vseld, my_mpu_opps[i].vsel, my_mpu_opps[i].x0, my_mpu_opps[i].x1, my_mpu_opps[i].x2, my_mpu_opps[i].x3, my_mpu_opps[i].x4, my_mpu_opps[i].x5, my_mpu_opps[i].x6, my_mpu_opps[i].x7, my_mpu_opps[i].x8, my_mpu_opps[i].x9); 		}

	return ret;
}

static int proc_mpu_opps_write(struct file *filp, const char __user *buffer,
		unsigned long len, void *data)
{
	uint index, rate, vsel;

	if(!len || len >= BUF_SIZE)
		return -ENOSPC;
	if(copy_from_user(buf, buffer, len))
		return -EFAULT;
	buf[len] = 0;
	if(sscanf(buf, "%d %d %d", &index, &rate, &vsel) == 3) {
		my_mpu_opps[index].rate = rate;
		my_mpu_opps[index].vsel = vsel;
	} else
		printk(KERN_INFO "overclock: insufficient parameters for mpu_opps\n");

	return len;
}                        
                        
static int proc_version_read(char *buffer, char **buffer_location,
		off_t offset, int count, int *eof, void *data)
{
	int ret;
	
	if (offset > 0)
		ret = 0;
	else
		ret = scnprintf(buffer, count, "%s\n", DRIVER_VERSION);

	return ret;
}

static int __init overclock_init(void)
{
	struct proc_dir_entry *proc_entry;

	printk(KERN_INFO "overclock: %s version %s\n", DRIVER_DESCRIPTION, DRIVER_VERSION);
	printk(KERN_INFO "overclock: by %s\n", DRIVER_AUTHOR);

	if(omap2_clk_init_cpufreq_table_addr)
		omap2_find_addrs();

	freq_table = (void *)freq_table_addr;
	my_mpu_opps = *(struct omap_opp2 **)mpu_opps_addr;
	policy = cpufreq_cpu_get(0);
	mpu_clk = clk_get(NULL, MPU_CLK);

	if(max_rate != DRIVER_DEFAULT_RATE || max_vsel != DRIVER_DEFAULT_VSEL)
		set_max_speed();

	buf = (char *)vmalloc(BUF_SIZE);

	proc_mkdir("overclock", NULL);
	proc_entry = create_proc_read_entry("overclock/info", 0444, NULL, proc_info_read, NULL);
	proc_entry = create_proc_read_entry("overclock/max_rate", 0644, NULL, proc_max_rate_read, NULL);
	proc_entry->write_proc = proc_max_rate_write;
	proc_entry = create_proc_read_entry("overclock/max_vsel", 0644, NULL, proc_max_vsel_read, NULL);
	proc_entry->write_proc = proc_max_vsel_write;
	proc_entry = create_proc_read_entry("overclock/freq_table_addr", 0644, NULL, proc_freq_table_addr_read, NULL);
	proc_entry->write_proc = proc_freq_table_addr_write;
	proc_entry = create_proc_read_entry("overclock/mpu_opps_addr", 0644, NULL, proc_mpu_opps_addr_read, NULL);
	proc_entry->write_proc = proc_mpu_opps_addr_write;
	proc_entry = create_proc_read_entry("overclock/omap2_clk_init_cpufreq_table_addr", 0644, NULL, proc_omap2_clk_init_cpufreq_table_addr_read, NULL);
	proc_entry->write_proc = proc_omap2_clk_init_cpufreq_table_addr_write;
	proc_entry = create_proc_read_entry("overclock/freq_table", 0644, NULL, proc_freq_table_read, NULL);
	proc_entry->write_proc = proc_freq_table_write;
	proc_entry = create_proc_read_entry("overclock/mpu_opps", 0644, NULL, proc_mpu_opps_read, NULL);
	proc_entry->write_proc = proc_mpu_opps_write;
	proc_entry = create_proc_read_entry("overclock/version", 0444, NULL, proc_version_read, NULL);

	return 0;
}

static void __exit overclock_exit(void)
{
	remove_proc_entry("overclock/version", NULL);
	remove_proc_entry("overclock/mpu_opps", NULL);
	remove_proc_entry("overclock/freq_table", NULL);
	remove_proc_entry("overclock/omap2_clk_init_cpufreq_table_addr", NULL);
	remove_proc_entry("overclock/mpu_opps_addr", NULL);
	remove_proc_entry("overclock/freq_table_addr", NULL);
	remove_proc_entry("overclock/max_vsel", NULL);
	remove_proc_entry("overclock/max_rate", NULL);
	remove_proc_entry("overclock/info", NULL);
	remove_proc_entry("overclock", NULL);
	 
	vfree(buf);

	if(max_rate != DRIVER_DEFAULT_RATE || max_vsel != DRIVER_DEFAULT_VSEL) {
		max_rate = DRIVER_DEFAULT_RATE;
		max_vsel = DRIVER_DEFAULT_VSEL;
		set_max_speed();
	}

	printk(KERN_INFO "overclock: removed overclocking and unloaded\n");
}

module_init(overclock_init);
module_exit(overclock_exit);

